#include <xil_exception.h>
#include <xil_io.h>
#include <xstatus.h>
#include <libeswpm/test_aux.h>
#include <xtime_l.h>
#include "pm_api_sys.h"
#include "pm_client.h"
#include "time.h"
#include "sleep.h"

#define MAX_COUNT	5

static XScuGic GicInst;
static XIpiPsu IpiInst;

#ifdef __aarch64__
#define TARGET_NODE		NODE_RPU_0
#else
#define TARGET_NODE		NODE_APU_0
#endif

/* Define mask encoding all target masters involved in the scenario */
#ifdef __aarch64__
#define OTHER_MASTERS_MASK	0x100
#elif __arm__
#define OTHER_MASTERS_MASK	IPI_PMU_0_TRIG_APU_MASK
#endif

static u32 request_suspend( u32 *latency) {
	XTime tStart, tEnd;
	u32 status;

	XTime_GetTime(&tStart);
	status = XPm_RequestSuspend(TARGET_NODE, REQUEST_ACK_NON_BLOCKING,
				    MAX_LATENCY, 0);
	IpiWaitForAck();
	XTime_GetTime(&tEnd);

	if (status != XST_SUCCESS)
		pm_print("Incorrect status returned for Request Suspend %s : %d\r\n",
				SHOW_DEFINE(TARGET_NODE), status);
	else
		*latency = (tEnd - tStart) / (COUNTS_PER_SECOND / 1000000);

	return status;
}

static u32 request_wakeup(u32 *latency, u32 *pu0_latency) {
	XTime tStart, tEnd, tpu0_up;
	u32 status;

	XTime_GetTime(&tStart);
	status = XPm_RequestWakeUp(TARGET_NODE, 0, 0, REQUEST_ACK_NON_BLOCKING);
	IpiWaitForAck();
	XTime_GetTime(&tpu0_up);
	SyncWaitForReady(OTHER_MASTERS_MASK);
	XTime_GetTime(&tEnd);
	SyncClearReady(OTHER_MASTERS_MASK);


	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Request wakeup %s : %d\r\n",
				SHOW_DEFINE(TARGET_NODE), status);
	} else {
		*pu0_latency = (tpu0_up - tStart) / (COUNTS_PER_SECOND / 1000000);
		*latency = (tEnd - tStart) / (COUNTS_PER_SECOND / 1000000);
	}
	return status;
}

static u32 measure_latency(void)
{
	u32 itteration, status, latency, pu0_latency;
	u32 susp_min = ~0, susp_max = 0, susp_avg = 0;
	u32 wake_min = ~0, wake_max = 0, wake_avg = 0;
	u32 pu0_wake_min = ~0, pu0_wake_max = 0, pu0_wake_avg = 0;
	u64 susp_total = 0, wake_total = 0, pu0_wake_total = 0;

	for (itteration = 0; itteration < MAX_COUNT; itteration++) {

		/* Measeure suspend latency */
		status = request_suspend(&latency);
		if (status != XST_SUCCESS)
			return XST_FAILURE;

		susp_min = (latency < susp_min ) ? latency : susp_min;
		susp_max = (latency > susp_max ) ? latency : susp_max;
		susp_total += latency;

		pm_print("Suspended in %ld.\r\n", latency);

		sleep(3);

		/* Measeure wakeup */
		status = request_wakeup(&latency, &pu0_latency);
		if (status != XST_SUCCESS)
			return XST_FAILURE;

#if !defined(__aarch64__)
		pm_print("APU0 powered up in %ld usec.\r\n", pu0_latency);
#endif
		pm_print("Resumed in %ld usec.\r\n", latency);
		wake_min = (latency < wake_min ) ? latency : wake_min;
		wake_max = (latency > wake_max ) ? latency : wake_max;
		wake_total += latency;

		pu0_wake_min = (pu0_latency < pu0_wake_min ) ? pu0_latency : pu0_wake_min;
		pu0_wake_max = (pu0_latency > pu0_wake_max ) ? pu0_latency : pu0_wake_max;
		pu0_wake_total += pu0_latency;
		sleep(3);
	}

	susp_avg = susp_total / MAX_COUNT;
	wake_avg = wake_total / MAX_COUNT;
	pu0_wake_avg = pu0_wake_total / MAX_COUNT;

	pm_print("Request Suspend Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), susp_min, susp_max, susp_avg);
#if !defined(__aarch64__)
	pm_print("APU0 Wakeup Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), pu0_wake_min, pu0_wake_max, pu0_wake_avg);
#endif
	pm_print("Wakeup Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), wake_min, wake_max, wake_avg);

	return XST_SUCCESS;
}

int main(void)
{
	u32 status;
	int ret = XST_SUCCESS;

	/* Initialize GIC, IPIs, and Xilpm */
	PmTestInit(&GicInst, &IpiInst);

	pm_print("INITIAL BOOT\r\n");

	/* Block until other masters in the system become ready */
	SyncWaitForReady(OTHER_MASTERS_MASK);
	SyncClearReady(OTHER_MASTERS_MASK);

#if !defined(__aarch64__)
        pm_print("Latency measurement while FPD on\r\n");
        /* Request NODE_GPU, So FPD will be on*/
        status = XPm_RequestNode(NODE_GPU, PM_CAP_ACCESS, MAX_QOS, REQUEST_ACK_BLOCKING);
        if (status != XST_SUCCESS && status != XST_PM_DOUBLE_REQ) {
                pm_print("request node failed\r\n");
                XIL_FAIL(1);
                exit(status);
        }

        status = measure_latency();
        if (status != XST_SUCCESS) {
                pm_print("latency measurement failed for %s\r\n", SHOW_DEFINE(TARGET_NODE));
                ret = status;
        }

        /* Relase NODE_GPU, So FPD will be off*/
        status = XPm_ReleaseNode(NODE_GPU);
        if (status != XST_SUCCESS && status != XST_PM_DOUBLE_REQ) {
                pm_print("release node failed\r\n");
                XIL_FAIL(1);
                exit(status);
        }

        pm_print("Latency measurement while FPD off\r\n");
#endif
	status = measure_latency();
	if (status != XST_SUCCESS) {
		pm_print("latency measurement failed for %s\r\n", SHOW_DEFINE(TARGET_NODE));
		ret = status;
	}

	pm_print("Latency measurement done\r\n");

	if (ret != XST_SUCCESS)
		pm_print("REGRESSION TEST FAIL\r\n");
	else
		pm_print("REGRESSION TEST PASS\r\n");

	return ret;
}
